#include <cmath>
#include <limits>
#include "nwt/startup.h"
#include "nwt/nucwin.h"
#include "3deng_dx8/3deng.h"
#include "timing.h"
#include "go.h"

#define AI_SAMPLES		1000
#define AI_PASS_THRES	2.0f

int ScreenX;
int ScreenY;

Widget *win;
Engine3D eng3d;
GraphicsContext *gc;
Timer timer;

Scene *scene;
Object *Board, *Grid, *White, *Black, *Table, *gui;
Camera cam;
PointLight light;

bool Init();
void MainLoop();
void CleanUp();
int KeyHandler(Widget *win, int key);
int CloseHandler(Widget *win, int arg);
int MouseHandler(Widget *win, int x, int y, bool left, bool middle, bool right);
int WheelHandler(Widget *win, int x, int y, int rot);

////////////// go specific shit //////////////
float GridSize = 5.5565f;
float height;

Vector3 GridPosition[19][19];
Matrix4x4 InvViewMatrix;		// for backprojection
Sphere bv[19][19];

std::list<Vector3> ExplosionPoints;
ParticleSystem **explosions;
int ExplosionCount;

bool explode = false;
float ExplosionStart;

int WhosTurn = BlackStone;
bool last_passed = false;

bool east_egg = true;

void InitializeGo();
void DrawBoard();
void SetupExplosionParticleSystem(ParticleSystem *psys, const Vector3 &pos);

int main() {

	win = NWCreateWindow("Go Game by John Tsiombikas (Nuclear / the Lab) 2003", NWT_CENTERX, NWT_CENTERY, 100, 50, 0);
	SetMainLoopFunc(MainLoop);
	SetHandler(HANDLER_KEY, KeyHandler);
	SetHandler(HANDLER_CLOSE, CloseHandler);
	SetHandler(HANDLER_MOUSE, (HandlerFunctionPtr)MouseHandler);
	SetHandler(HANDLER_WHEEL, (HandlerFunctionPtr)WheelHandler);

	if(!Init()) return 0;

	return NWMainLoop(RealTimeLoop);
}

bool Init() {

	ContextInitParameters cip;
	try {
		cip = eng3d.LoadContextParamsConfigFile("n3dinit.conf");
		gc = eng3d.CreateGraphicsContext(win, 0, &cip);
	}
	catch(EngineInitException except) {
		MessageBox(win, except.GetReason().c_str(), "Fatal Error", MB_OK | MB_ICONSTOP);
		return false;
	}
	ScreenX = cip.x;
	ScreenY = cip.y;
	NWResize(win, ScreenX, ScreenY);
	NWResizeClientArea(win, WS_OVERLAPPEDWINDOW);

	cam.SetCamera(Vector3(0.0f, 100.0f, -100.0f), Vector3(0.0f, 0.0f, 0.0f), VECTOR3_J);

	light = PointLight(Vector3(0.0f, 100.0f, 0.0f));

	gc->SetAmbientLight(Color(0.1f));
	
	SceneLoader::SetGraphicsContext(gc);
	SceneLoader::SetDataPath("data/");
	SceneLoader::LoadObject("data/geometry2.3ds", "Board", &Board);
	SceneLoader::LoadObject("data/geometry2.3ds", "Grid", &Grid);
	SceneLoader::LoadObject("data/geometry2.3ds", "White", &White);
	SceneLoader::LoadObject("data/geometry2.3ds", "Black", &Black);
	SceneLoader::LoadObject("data/geometry2.3ds", "Trapezi", &Table);

	gui = new Object(gc);
	gui->CreatePlane(1.6f, 0);
	gui->material = Material(1.0f, 1.0f, 1.0f);
	gui->material.Alpha = 0.99f;
	gui->material.SetTexture(gc->texman->LoadTexture("data/gui01.jpg"), TextureMap);
	gui->SetScaling(2.0f, 1.0f, 1.0f);
	gui->SetTranslation(-6.0f, 4.0f, 0.0f);
	gui->SetBlendFunc(BLEND_ONE, BLEND_ONE);

	InitializeGo();
	
	Matrix4x4 ScaleMat;
	ScaleMat.SetScaling(1.0f, 0.4f, 1.0f);
	Vertex *varray1 = Black->GetTriMesh()->GetModVertexArray();
	Vertex *varray2 = White->GetTriMesh()->GetModVertexArray();
	for(dword i=0; i<Black->GetTriMesh()->GetVertexCount(); i++) {
		varray1[i].pos.Transform(ScaleMat);
		varray2[i].pos.Transform(ScaleMat);
	}

	timer.Start();

	return true;
}

void MainLoop() {
	float t = (float)timer.GetMilliSec() / 1000.0f;

	// fade in
	if(t < 3.0f) {
		light.SetIntensity(t / 3.0f);
	} else {
		light.SetIntensity(1.0f);
	}
	/////////

	cam.CreateCameraMatrix();
	InvViewMatrix = cam.GetCameraMatrix().Inverse();
	gc->SetViewMatrix(cam.GetCameraMatrix());
	light.SetLight(0, gc);


	gc->Clear(0);
	gc->ClearZBufferStencil(1.0f, 0);

	gc->SetZBuffering(true);
	gc->SetZWrite(true);
	gc->SetZFunc(CMP_LEQUAL);

	Board->Render();
	Grid->Render();
	Table->Render();	
	
	DrawBoard();

	// if we removed any string make a funky explosion
	if(explode) {
		for(int i=0; i<ExplosionCount; i++) {
			explosions[i]->Update(t - ExplosionStart);
			explosions[i]->Render();
		}
		if(!explosions[0]->CountParticles()) {
			ExplosionPoints.erase(ExplosionPoints.begin(), ExplosionPoints.end());
			explode = false;
		}

	}
	
	// "who's turn is it" indicator
	gc->SetZBuffering(false);

	Matrix4x4 ViewMat;
	ViewMat.Translate(0.0f, 0.0f, 10.0f);
	gc->SetViewMatrix(ViewMat);
	
	gc->SetLighting(false);
	gui->Render();
	gc->SetLighting(true);

	
	Camera TempCam;
	TempCam.SetCamera(Vector3(0.0f, 80.0f, -80.0f), Vector3(65.0f, -260.0f, 0.0f), VECTOR3_J);
	TempCam.CreateCameraMatrix();
	gc->SetViewMatrix(TempCam.GetCameraMatrix());

	Object *indicator = WhosTurn == BlackStone ? Black : White;
	indicator->ResetTransform();
	indicator->Rotate(t, t, 0.0f);
	indicator->Render();
	indicator->ResetTransform();

	gc->SetZBuffering(true);
	
	gc->Flip();
}

void CleanUp() {

	if(explosions) {
		for(int i=0; i<ExplosionCount; i++) {
			if(explosions[i]) delete explosions[i];
		}
	}

	delete Board;
	delete Grid;
	delete White;
	delete Black;

}

///////////////////////////////////////////////////////////

void InitializeGo() {
	// find the y pos of the stones
	Vector3 pos(0,0,0);
	pos.Transform(White->GetWorldTransform());
	height = pos.y;

	// find the grid size
	GridSize = (Grid->GetTriMesh()->GetVertexArray()[0].pos - Grid->GetTriMesh()->GetVertexArray()[1].pos).Length();

	// initialize the board
	memset(GoBoard, Empty, 19*19*sizeof(GoNode));

	// initialize the grid world positions array
	const Vertex *varray = Grid->GetTriMesh()->GetVertexArray();
	const Vertex *ptr = varray;

	for(int x=0; x<19; x++) {
		for(int y=0; y<19; y++) {
			GridPosition[x][y] = (*ptr++).pos;
			bv[x][y].center = GridPosition[x][y];
			bv[x][y].radius = GridSize / 2.0f;			
		}
	}

	for(int i=0; i<19; i++) {
		for(int j=0; j<19; j++) {
			GoBoard[i][j].AdjNodes[0] = j > 0 ? &GoBoard[i][j-1] : 0;
			GoBoard[i][j].AdjNodes[1] = i < 18 ? &GoBoard[i+1][j] : 0;
			GoBoard[i][j].AdjNodes[2] = j < 18 ? &GoBoard[i][j+1] : 0;
			GoBoard[i][j].AdjNodes[3] = i > 0 ? &GoBoard[i-1][j] : 0;
			GoBoard[i][j].bvol = &bv[i][j];
		}
	}
}

void DrawBoard() {
	for(int x=0; x<19; x++) {
		for(int y=0; y<19; y++) {
			if(GoBoard[x][y].state != Empty) {

				if(GoBoard[x][y].state == BlackStone) {
					Black->SetTranslation(GridPosition[x][y].x, height, GridPosition[x][y].z);
					Black->Render();
				} else {
					White->SetTranslation(GridPosition[x][y].x, height, GridPosition[x][y].z);
					White->Render();
				}
			}
		}
	}
}

const float Aspect = 1.333333333f;
const float HalfFOVTan = tanf(QuarterPi/2.0f);

POINT PickGridPoint(int x, int y) {
	
	// backproject to world space
	float NearPlane, FarPlane;
	cam.GetClippingPlanes(&NearPlane, &FarPlane);

	// find normalized screen coords
	Vector3 pos(((float)x / (float)(ScreenX>>1) - 1.0f) * Aspect, 1.0f - (float)y /(float)(ScreenY>>1), NearPlane);

	// find position in view space 
	pos.x *= HalfFOVTan;
	pos.y *= HalfFOVTan;

	// find the coordinates of the ray on the clipping planes
	Vector3 p1 = Vector3(pos.x * NearPlane, pos.y * NearPlane, NearPlane);
	Vector3 p2 = Vector3(pos.x * FarPlane, pos.y * FarPlane, FarPlane);

	// transform the ray endpoints to world coordinates
	p1.Transform(InvViewMatrix);
	p2.Transform(InvViewMatrix);

	// test for intersections
	Ray ray;
	ray.Origin = p1;
	ray.Direction = p2 - p1;
	float mag = ray.Direction.Length();
	ray.Direction.Normalize();


	float ClosestT = FLT_MAX;
	POINT GridPoint;
	GridPoint.x = -1;
	GridPoint.y = -1;
	for(int x=0; x<19; x++) {
		for(int y=0; y<19; y++) {
			float t = bv[x][y].TestRayIntersection(ray, true, mag);
			if(t != -1.0f && t < ClosestT) {
				ClosestT = t;
				GridPoint.x = x;
				GridPoint.y = y;
			}
		}
	}

	return GridPoint;
}

void MoveEffects(int x, int y) {

	bool new_captures = false;

	// check if there is a capture
	if(x < 18 && !CountStringFreedoms(&GoBoard[x+1][y])) {
		ExplosionPoints.erase(ExplosionPoints.begin(), ExplosionPoints.end());
		RemoveString(&GoBoard[x+1][y]);
		explode = true;
		new_captures = true;
	}
	if(y < 18 && !CountStringFreedoms(&GoBoard[x][y+1])) {
		ExplosionPoints.erase(ExplosionPoints.begin(), ExplosionPoints.end());
		RemoveString(&GoBoard[x][y+1]);
		explode = true;
		new_captures = true;
	}
	if(x > 0 && !CountStringFreedoms(&GoBoard[x-1][y])) {
		ExplosionPoints.erase(ExplosionPoints.begin(), ExplosionPoints.end());
		RemoveString(&GoBoard[x-1][y]);
		explode = true;
		new_captures = true;
	}
	if(y > 0 && !CountStringFreedoms(&GoBoard[x][y-1])) {
		ExplosionPoints.erase(ExplosionPoints.begin(), ExplosionPoints.end());
		RemoveString(&GoBoard[x][y-1]);
		explode = true;
		new_captures = true;
	}

	if(explode && new_captures) {
		ExplosionStart = (float)timer.GetMilliSec() / 1000.0f;

		if(explosions) {
			for(int i=0; i<ExplosionCount; i++) {
				if(explosions[i]) delete explosions[i];
			}
		}
		ExplosionCount = (int)ExplosionPoints.size();
		explosions = new ParticleSystem*[ExplosionCount];
		
		std::list<Vector3>::iterator iter = ExplosionPoints.begin();
		for(int i=0; i<ExplosionCount; i++) {
			explosions[i] = new ParticleSystem(gc);
			SetupExplosionParticleSystem(explosions[i], (*iter++));
		}					
	}
}

void EndGame() {
    MessageBox(0, "End game", "game ended", MB_OK);
}

///////// handlers /////////
int KeyHandler(Widget *win, int key) {

	static char buffer[8] = {0,0,0,0,0,0,0,0};
	for(int i=0; i<6; i++) {
		buffer[i] = buffer[i+1];
	}
	buffer[6] = (char)key;

	switch(key) {
	case VK_ESCAPE:
		NWCloseWindow(win);
		break;

	case 'a':
	case 'A':
		//cam.Zoom(0.05f);
		break;

	case 'z':
	case 'Z':
		//cam.Zoom(-0.05f);
		break;

	case '1':
		east_egg = (east_egg + 1) % 2;
		break;

	case 'p':
	case 'P':
		if(last_passed) EndGame();
		WhosTurn = (WhosTurn % 2) + 1;
		last_passed = true;
		break;

	}

	if(!strcmp(buffer, "nuclear") || !strcmp(buffer, "NUCLEAR")) foo(east_egg ? 1 : 0);
	
	return 0;
}

int CloseHandler(Widget *win, int arg) {
	CleanUp();
	return 0;
}

int WheelHandler(Widget *win, int x, int y, int rot) {
	float tr = rot > 0 ? 0.05f : -0.05f;

	cam.Zoom(tr);
	
	return 0;
}

float t = 0.0f;

int MouseHandler(Widget *win, int x, int y, bool left, bool middle, bool right) {
	static POINT PrevPos;
	POINT MoveDiff;

	if(left) {
		POINT pt = PickGridPoint(x, y);
		if(pt.x != -1 && pt.y != -1) {
			if(!GoBoard[pt.x][pt.y].state) {
				if(IllegalMove(&GoBoard[pt.x][pt.y], WhosTurn)) {
					MessageBox(win, "Feeling suicidal today huh? :)", "Illegal move", MB_OK | MB_ICONEXCLAMATION);
					return 0;
				}

				GoBoard[pt.x][pt.y].state = WhosTurn;

				WhosTurn = (WhosTurn % 2) + 1;
				last_passed = false;

				MoveEffects(pt.x, pt.y);				
			}
		}

		if(WhosTurn == AIPlayer) {
			int move_x, move_y;
			bool AI_moved = MakeMove(&move_x, &move_y, AI_SAMPLES, AI_PASS_THRES, MainLoop);
			if(!AI_moved) {
				if(last_passed) EndGame();
				last_passed = true;
			} else {
				MoveEffects(move_x, move_y);
				last_passed = false;
			}
			WhosTurn = HumanPlayer;
		}

	}
	
	if(right) {
		MoveDiff.x = x - PrevPos.x;
		MoveDiff.y = y - PrevPos.y;
		float xangle = (float)MoveDiff.x / 50.0f;
		float dy = (float)MoveDiff.y;

		cam.Rotate(0.0f, xangle, 0.0f);
		cam.SetPosition(cam.GetPosition() + Vector3(0.0f, dy, 0.0f));
	}

	if(middle) {
		POINT pt = PickGridPoint(x, y);
		if(pt.x != -1 && pt.y != -1) {
			float eval = EvaluatePosition(&GoBoard[pt.x][pt.y], WhosTurn);
			char eval_str[100];

			sprintf(eval_str, "Position Evaluation: %f", eval);
			SetWindowText(win, eval_str);
		}
	}

	PrevPos.x = x;
	PrevPos.y = y;

	return 0;
}


////////////////////////////////////////////////////

void SetupExplosionParticleSystem(ParticleSystem *psys, const Vector3 &pos) {
	psys->SetPosition(pos);
	psys->SetGravitualForce(0.9f);
	psys->SetFriction(0.05f);
	psys->SetInitialColor(1.0f, 0.3f, 0.1f);
	psys->SetDeathColor(0.0f, 0.0f, 0.0f);
	psys->SetParticleSize(7.0f);
	psys->SetParticleLife(60);
	psys->SetSpawnRadius(3.0f);
	psys->SetSpawnRate(25.0f);
	psys->SetTexture(gc->texman->LoadTexture("STOCKTEX_BLOB"));
	psys->SetShootDirection(Vector3(0.0f, 3.0f, 0.0f));
	psys->SetMaxDispersionAngle(QuarterPi / 2.0f);
	psys->SetBlendingMode(BLEND_ONE, BLEND_ONE);
	psys->SetSpawnRateChange(-2);
}